// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020 Florian Schmaus
#pragma once

#include <cassert>			// for assert
#include <cstdint>			// for uintptr_t
#include <functional>		// for function
#include <iosfwd>				// for ostream
#include <type_traits>	// for remove_reference<>::type
#include <utility>

#include "Common.hpp"	 // for func_t, DIE, ALIGN_TO_CACHE_LINE
#include "Debug.hpp"	 // for LOGD, LogSubsystem, LogSubsystem::C, Logger

#ifndef NDEBUG
#include <cstring>	// for memset
#endif

extern "C" [[noreturn]] void switch_and_load_context(void** toTos);
// *Not* marked as 'noreturn' because save_and_switch_context does
// *actually return at some point, unlike the other switch_*
// *functions.
extern "C" void save_and_switch_context(void** toTos, void** fromTos);
extern "C" [[noreturn]] void switch_context(void** toTos);

class ALIGN_TO_CACHE_LINE Context : Logger<LogSubsystem::C> {
 private:
	static const unsigned int CONTEXT_SIZE = 0xffff;	// 1024 * 1024 * 4;

	static thread_local Context* currentContext;

	void* const tos;

	//	unsigned valgrindStackId;
	void* savedStackpointer;

	const void* alphaFunctionIpLocation;

	const func_t mainFunction;

	func_t startAndResumeHook;

	// Allign the stack to a cache line, so that it isn't shared with
	// the other members of this class.
	// NOLINTNEXTLINE(modernize-avoid-c-arrays)
	ALIGN_TO_CACHE_LINE char context[CONTEXT_SIZE];

	friend auto operator<<(std::ostream&, const Context&) -> std::ostream&;

	/**
	 * The first function that a newly started context will
	 * execute. This function simply first runs the pre main hook and
	 * then the main function.
	 */
	static void kickoff() {
		Context* context = getCurrentContext();
		context->startAndResumeHook();
		context->mainFunction();
	}

 public:
	// cppcheck-suppress noExplicitConstructor selfInitialization
	Context(func_t mainFunction)
			// Align the Top-of-Stack (tos) to 16 byte.
			: tos((void*)((uintptr_t)(context + CONTEXT_SIZE) & (~0xf))),
				mainFunction(std::move(std::move(mainFunction))) {
		//		valgrindStackId = VALGRIND_STACK_REGISTER(context, context + CONTEXT_SI);

#ifndef NDEBUG
		// Write the stack full of 0xcc bytes, which just happen to be
		// the 'int3' instruction, which will trigger a breakpoint
		// when executed.
		void* res = memset(context, 0xcc, (char*)tos - context);
		if (!res) DIE;
		// Mark the last valid 16 bytes just below the top of stack.
		res = memset(((uintptr_t*)tos) - 1, 0xab, sizeof(uintptr_t));
		if (!res) DIE;
		// Mark the eventually existing unused top stack bytes
		res = memset(tos, 0xee, &context[CONTEXT_SIZE] - (char*)tos);
		if (!res) DIE;
#endif

		setEmptyHook();

		// We write the real kickoff function's address at
		// TOS - 2 so it's called after this context has been actived.
		// This also ensures that the stack is correctly aligned for
		// SSE operations, which require the stack to be 8-byte, but
		// not 16-byte aligned.
		alphaFunctionIpLocation = savedStackpointer = (uintptr_t*)tos - 2;

		void** alphaSavedIp = reinterpret_cast<void**>(savedStackpointer);
		void (*f)() = &kickoff;
		assert(f != nullptr);
		*(alphaSavedIp) = reinterpret_cast<void*>(f);
	}

	// NOLINTNEXTLINE(modernize-use-equals-default)
	~Context() {
		//		VALGRIND_STACK_DEREGISTER(valgrindStackId);
	}

	inline void setEmptyHook() {
		startAndResumeHook = []() {};
	}

	inline void setHook(func_t hook) { startAndResumeHook = std::move(hook); }

	[[nodiscard]] inline auto getTos() const -> const void* { return tos; }

	/**
	 * Start this context.
	 */
	[[noreturn]] inline void start() {
		LOGD("starting");
		currentContext = this;
		switch_context(&savedStackpointer);
	}

	/**
	 * Save the current context and switch to a new one. This method is
	 * usally used when a synchronization primitve decided that it is
	 * time to block the current execution context. In such a case
	 * 'newContext' is a free context.
	 */
	inline void saveAndStart(Context* newContext) {
		LOGD("saving and switching to " << *newContext);
		currentContext = newContext;

		// Save the current context and start a new one.
		save_and_switch_context(&(newContext->savedStackpointer), &savedStackpointer);
		// If we are here, then the context was unblocked.

		// Call the hook again, as it may be set by the context which
		// activated us.
		startAndResumeHook();
	}

	/**
	 * Discard the current context and restore the given one. This
	 * method is usually used when a synchronization primitive decided
	 * that it is time to unblock a context. In such a case
	 * 'newContext' is the context that became runable again. Note
	 * that this method is usually called by a so called "unblock
	 * Fiber". This method does not return.
	 */
	// The discardAndResume() method could theoretically be made static in
	// non-debug builds.
	// NOLINTNEXTLINE(readability-convert-member-functions-to-static)
	[[noreturn]] inline void discardAndResume(Context* context) {
		//		assert(context != nullptr);

		LOGD("discarding and switching to " << context);
		currentContext = context;
		switch_and_load_context(&(context->savedStackpointer));
	}

	inline static auto getCurrentContext() -> Context* { return currentContext; }
};
