// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020 Florian Schmaus
#include "ContextManager.hpp"

#include <cassert>	// for assert
#include <new>			// for operator new, operator delete
#include <utility>

#include "Context.hpp"		 // for Context
#include "Debug.hpp"			 // for LOGD
#include "Dispatcher.hpp"	 // for Dispatcher
#include "Runtime.hpp"		 // for Runtime

ContextManager::ContextManager(Runtime& runtime) : MemoryManager(runtime), runtime(runtime) {
	auto newWorkerHook = [this]() {
		for (unsigned int i = 0; i < CONTEXT_MANAGER_FIRST_LAYER_QUEUE_SIZE * 2; ++i) {
			auto* context = new Context(this->runtime.dispatcher.getDispatchLoop());
			putFreeContext(context);
		}
	};
	// Note that it is important that this hook is executed *after*
	// the one of the MemoryManager superclass.
	runtime.addNewWorkerHook(newWorkerHook);
}

auto ContextManager::getFreeContext() -> Context* {
#ifdef CM_WITH_MEMORY_MANAGER
	bool malloced;
	void* memory = getMemory(&malloced);
	if (malloced) {
		return new (memory) Context(this->runtime.dispatcher.getDispatchLoop());
	}
	return static_cast<Context*>(memory);
#else
	return new Context(this->runtime.dispatcher.getDispatchLoop());
#endif
}

void ContextManager::putFreeContext(Context* context) {
#ifdef CM_WITH_MEMORY_MANAGER
	putMemory(context);
#else
	delete context;
#endif
}

void ContextManager::start() {
	Context* freeContext = getFreeContext();
	freeContext->start();
}

/**
 * Save the current context and start a new one.
 */
void ContextManager::saveAndStartNew(func_t freshContextHook) {
	Context* freeContext = getFreeContext();
	freeContext->setHook(std::move(freshContextHook));
	Context::getCurrentContext()->saveAndStart(freeContext);
}

#ifdef NDEBUG
#define HOOK_CAPTURE_LIST contextToFree
#else
// We only need to caputure 'this' for the LOGD() debug macro.
#define HOOK_CAPTURE_LIST this, contextToFree
#endif

/**
 * Discard the current context and resume a saved one.
 */
// The discardAndResume() method could theoretically be made static in
// non-debug builds.
// NOLINTNEXTLINE(readability-convert-member-functions-to-static)
void ContextManager::discardAndResume(Context* context) {
	assert(context != nullptr);

	Context* contextToFree = Context::getCurrentContext();
	context->setHook([HOOK_CAPTURE_LIST] {
		LOGD("Freeing context " << contextToFree);
		putFreeContext(contextToFree);
	});
	contextToFree->discardAndResume(context);
}
