// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020 Florian Schmaus
#include "Runtime.hpp"

#include <pthread.h>	// for pthread_t, pthread_attr_init

#include <cerrno>	 // for errno
// Non portable.
#include <sched.h>				// for cpu_set_t, CPU_SET, CPU_ZERO
#include <sys/sysinfo.h>	// for get_nprocs

#include <cstdlib>	// for rand, srand, abort
#include <memory>		// for __shared_ptr_access, shared_ptr

#include "Common.hpp"					 // for DIE_MSG_ERRNO, DIE, DIE_MSG
#include "ContextManager.hpp"	 // for ContextManager
#include "Debug.hpp"					 // for DBG, ABORT, LOGD, LOGE
#include "Dispatcher.hpp"			 // for Dispatcher
#include "Emper.hpp"
#include "Fiber.hpp"										 // for Fiber
#include "RuntimeStrategy.hpp"					 // for RuntimeStrategy
#include "RuntimeStrategyStats.hpp"			 // for RuntimeStrategyStats
#include "strategies/ws/WsStrategy.hpp"	 // for WsStrategy, WsStrategy::INST...

#ifndef NDEBUG
#include <syscall.h>	// for SYS_gettid
#include <unistd.h>		// for syscall

#include <ostream>	// for operator<<, basic_ostream<>:...
#endif

std::mutex Runtime::currentRuntimeMutex;
Runtime* Runtime::currentRuntime;
thread_local unsigned int Runtime::seed;
thread_local workerid_t Runtime::workerId;
RuntimeStrategy& Runtime::DEFAULT_STRATEGY = WsStrategy::INSTANCE;

Runtime::Runtime(workerid_t workerCount, RuntimeStrategy& strategy, unsigned int seed)
		: workerCount(workerCount),
			workerLatch(workerCount),
			strategy(strategy),
			scheduler(strategy.getScheduler(*this)),
			dispatcher(strategy.getDispatcher(*this)),
			contextManager(*(new ContextManager(*this))),
			randomEngine(seed),
			atLeastOneWorkerIsSleeping(false) {
	threads = new pthread_t[workerCount];
	workerIds = new workerid_t[workerCount];

	const int nprocs = get_nprocs();

	{
		std::lock_guard<std::mutex> lock(currentRuntimeMutex);
		if (currentRuntime != nullptr) {
			DIE_MSG("There is already a runtime active");
		}
		currentRuntime = this;
	}

	for (workerid_t i = 0; i < workerCount; ++i) {
		pthread_attr_t attr;
		errno = pthread_attr_init(&attr);
		if (errno) DIE;

		// Start non portable.
		cpu_set_t cpuset;
		CPU_ZERO(&cpuset);
		CPU_SET(i % nprocs, &cpuset);

		errno = pthread_attr_setaffinity_np(&attr, sizeof(cpuset), &cpuset);
		if (errno) DIE_MSG_ERRNO("pthread_attr_setaffinity_np() failed");
		// End non portable.

		// Load the worker ID into the worker ID map.
		workerIds[i] = i;

		auto thread_function = [](void* voidWorkerId) -> void* {
			return currentRuntime->workerLoop(voidWorkerId);
		};
		errno = pthread_create(&threads[i], &attr, thread_function, &workerIds[i]);
		if (errno) DIE_MSG_ERRNO("pthread_create() failed");
	}

	if constexpr (emper::STATS) {
		int res = std::atexit(&printLastRuntimeStats);
		if (res) {
			DIE_MSG("could not register printStats() with at_exit()");
		}
	}
}

Runtime::~Runtime() {
	DBG("Runtime " << this << " is terminating");
	for (workerid_t i = 0; i < workerCount; ++i) {
		DBG("Runtime " << this << " is cancelling worker " << unsigned(i));
		errno = pthread_cancel(threads[i]);
		if (errno) {
			DIE_MSG_ERRNO("pthread_cancel() failed");
		}
	}
	{
		std::lock_guard<std::mutex> lock(currentRuntimeMutex);
		currentRuntime = nullptr;
	}
	DBG("Runtime " << this << " terminated");
}

auto Runtime::workerLoop(void* voidWorkerId) -> void* {
	workerId = *(workerid_t*)voidWorkerId;
	LOGD("Worker loop started by thread " << syscall(SYS_gettid));

	int oldType;
	errno = pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &oldType);
	if (errno) {
		DIE_MSG_ERRNO("pthread_setcanceltype() failed");
	}

	// Initialze the workers PRNG seed.
	seed = uniformIntDistribution(randomEngine);

	for (const auto& f : newWorkerHooks) f();

	workerLatch.count_down_and_wait();

	contextManager.start();

	abort();
	return nullptr;
}

auto Runtime::nextFiber() -> Fiber* { return scheduler.nextFiber(); }

void Runtime::waitUntilFinished() {
	for (workerid_t i = 0; i < workerCount; ++i) {
		pthread_join(threads[i], nullptr);
	}
}

void Runtime::printStats() {
	auto runtimeStrategyStats = strategy.getStats();
	runtimeStrategyStats->print();
}

void Runtime::printLastRuntimeStats() {
	std::lock_guard<std::mutex> lock(currentRuntimeMutex);
	if (currentRuntime == nullptr) {
		LOGE("Could not print current runtime stats");
		return;
	}

	currentRuntime->printStats();
}

auto Runtime::inRuntime() -> bool { return Dispatcher::isDispatchedControlFlow(); }

void Runtime::executeAndWait(std::function<void()> f) {
	if (inRuntime()) {
		ABORT("Ca not use executeAndWait() from within the Runtime");
	}

	std::mutex fiberFinished;
	fiberFinished.lock();

	Fiber* fiber = Fiber::from([&] {
		f();

		fiberFinished.unlock();
	});

	schedule(*fiber);

	fiberFinished.lock();
}
