// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020 Florian Schmaus
#pragma once

#include <pthread.h>	// for pthread_t

#include <atomic>							 // for atomic, memory_order_relaxed
#include <condition_variable>	 // for condition_variable
#include <cstddef>						 // for size_t
#include <cstdint>						 // for intptr_t
#include <functional>					 // for function
#include <mutex>							 // for mutex, lock_guard, unique_lock
#include <random>
#include <thread>	 // for thread
#include <vector>	 // for vector

#include "Common.hpp"					 // for ALIGN_TO_CACHE_LINE
#include "Debug.hpp"					 // for LogSubsystem, LogSubsystem::RUNTI, Logger
#include "Scheduler.hpp"			 // for Scheduler
#include "emper-common.h"			 // for workerid_t
#include "lib/sync/Latch.hpp"	 // for Latch

class ContextManager;
class Dispatcher;
class Fiber;
class RuntimeStrategy;

class Runtime : public Logger<LogSubsystem::RUNTI> {
 private:
	static std::mutex currentRuntimeMutex;
	static Runtime* currentRuntime;

	static thread_local unsigned int seed;
	static thread_local workerid_t workerId;

	const workerid_t workerCount;

	std::vector<std::function<void(void)>> newWorkerHooks;

	Latch workerLatch;

	RuntimeStrategy& strategy;
	Scheduler& scheduler;
	Dispatcher& dispatcher;
	ContextManager& contextManager;
	pthread_t* threads;
	workerid_t* workerIds;

	std::default_random_engine randomEngine;
	std::uniform_int_distribution<unsigned int> uniformIntDistribution;

	auto workerLoop(void* workerId) -> void*;

	std::mutex workerSleepMutex;
	std::condition_variable workerSleepConditionVariable;
	ALIGN_TO_CACHE_LINE std::atomic<bool> atLeastOneWorkerIsSleeping;

	static RuntimeStrategy& DEFAULT_STRATEGY;

	static void printLastRuntimeStats();

 protected:
	void addNewWorkerHook(const std::function<void(void)>& hook) { newWorkerHooks.push_back(hook); };

	inline void notifyAboutNewWork() {
		if (!atLeastOneWorkerIsSleeping.load(std::memory_order_relaxed)) return;

		std::lock_guard<std::mutex> lk(workerSleepMutex);
		workerSleepConditionVariable.notify_all();
	}

	void dispatcherLoopSleep() {
		std::unique_lock<std::mutex> lk(workerSleepMutex);
		atLeastOneWorkerIsSleeping.store(true, std::memory_order_relaxed);
		workerSleepConditionVariable.wait(lk);
		atLeastOneWorkerIsSleeping.store(false, std::memory_order_relaxed);
	}

 public:
	Runtime() : Runtime(std::thread::hardware_concurrency()) {}

	Runtime(workerid_t workerCount) : Runtime(workerCount, DEFAULT_STRATEGY) {}

	Runtime(RuntimeStrategy& strategy) : Runtime(std::thread::hardware_concurrency(), strategy) {}

	Runtime(workerid_t workerCount, RuntimeStrategy& strategy,
					unsigned int seed = std::random_device()());

	~Runtime();

	inline void schedule(Fiber& fiber) { scheduler.schedule(fiber); }

	auto nextFiber() -> Fiber*;

	// https://stackoverflow.com/a/3747462/194894
	static inline auto rand() -> int {
		seed = 214013 * seed + 2531011;
		auto shifted_seed = seed >> 16;
		return *reinterpret_cast<int*>(&shifted_seed) & 0x7FFF;
	}

	static inline auto getWorkerId() -> workerid_t { return workerId; }

	[[nodiscard]] inline auto getWorkerCount() const -> workerid_t { return workerCount; }

	static inline auto getRuntime() -> Runtime* { return currentRuntime; }

	inline auto getContextManager() -> ContextManager& { return contextManager; }

	inline auto getStrategy() -> RuntimeStrategy& { return strategy; }

	void waitUntilFinished();

	void printStats();

	static auto inRuntime() -> bool;

	void executeAndWait(std::function<void()> f);

	friend ContextManager;
	friend Scheduler;
	friend Dispatcher;
	template <typename T, intptr_t WS_QUEUE_SIZE, size_t WORKER_EXCLUSIVE_QUEUE_SIZE>
	friend class MemoryManager;
};
