// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020 Florian Schmaus
#include "LawsScheduler.hpp"

#include <atomic>	 // for atomic, memory_order_relaxed

#include "Common.hpp"
#include "Debug.hpp"
#include "Emper.hpp"
#include "LawsStrategy.hpp"	 // IWYU pragma: keep
#include "Runtime.hpp"

#define EMPER_OVERFLOW_QUEUE

thread_local LawsScheduler::LawsMpscQueue LawsScheduler::priorityQueue;

thread_local LawsScheduler::WsQueue<LawsScheduler::QUEUE_SIZE> LawsScheduler::queue;

LawsScheduler::LawsScheduler(Runtime& runtime, LawsStrategy& lawsStrategy)
		: Scheduler(runtime), lawsStrategy(lawsStrategy) {
	const workerid_t workerCount = runtime.getWorkerCount();
	queues = new LawsScheduler::WsQueue<QUEUE_SIZE>*[workerCount];
	priorityQueues = new LawsScheduler::LawsMpscQueue*[workerCount];
	mainThreadQueue = &queue;

	auto newWorkerHook = [this]() {
		workerid_t workerId = Runtime::getWorkerId();
		queues[workerId] = &queue;
		priorityQueues[workerId] = &priorityQueue;
	};
	addNewWorkerHook(newWorkerHook);
}

void LawsScheduler::schedule(Fiber& fiber) {
	LOGD("Scheduling fiber " << &fiber);

	workeraffinity_t* const affinity_buffer = getAffinityBuffer(fiber);
	if (affinity_buffer) {
		workeraffinity_t affinity = *affinity_buffer;
		workerid_t workerId = Runtime::getWorkerId();
		if (affinity == workerId) {
			goto scheduleToLocalWsQueue;
		}

		if (affinity == Fiber::NOT_AFFINE) {
			goto scheduleToLocalWsQueue;
		}

		// We found a fiber to schedule on a remote prority queue.
		increaseRefCount(fiber);
		priorityQueues[affinity]->enqueue(&fiber);
		if constexpr (emper::STATS) {
			lawsStrategy.scheduledFibersToRemotePriority.fetch_add(1, std::memory_order_relaxed);
		}
	}

scheduleToLocalWsQueue:
	bool pushed = queue.pushBottom(&fiber);
	if (unlikely(!pushed)) {
#ifdef EMPER_OVERFLOW_QUEUE
		priorityQueue.enqueue(&fiber);
#else
		ABORT("Could not push fiber " << &fiber << " into queue");
#endif
	} else {
		if constexpr (emper::STATS) {
			lawsStrategy.scheduledFibersToLocal.fetch_add(1, std::memory_order_relaxed);
		}
	}

	if constexpr (emper::WORKER_SLEEP) {
		notifyRuntimeAboutNewWork();
	}
}

auto LawsScheduler::nextFiber() -> Fiber* {
	Fiber* fiber = priorityQueue.dequeue();
	if (fiber != nullptr) {
		// We fetched a fiber from your local priority queue.
		if constexpr (emper::STATS) {
			auto flag = static_cast<unsigned int>(LawsStrategy::FiberSource::fromPriority);
			fiber->setFlag(flag);
		}

		return fiber;
	}

	bool poped = queue.popBottom(&fiber);

	if (likely(poped)) {
		if constexpr (emper::STATS) {
			auto flag = static_cast<unsigned int>(LawsStrategy::FiberSource::fromLocal);
			fiber->setFlag(flag);
		}

		return fiber;
	}

	const workerid_t workerCount = runtime.getWorkerCount();
	workerid_t startWorkerId = Runtime::rand() % workerCount;

	if (unlikely(startWorkerId == Runtime::getWorkerId())) startWorkerId++;
	workerid_t victim = startWorkerId % workerCount;

	poped = queues[victim]->popTop(&fiber);
	if (poped) {
		if (emper::STATS) {
			auto flag = static_cast<unsigned int>(LawsStrategy::FiberSource::stolen);
			fiber->setFlag(flag);
		}

		return fiber;
	}

	// TODO: Reduce pressure on mainThreadQueue by only checking every N-th time.

	// Try the main thread queue to get work as last resort.
	poped = mainThreadQueue->popTop(&fiber);
	if (poped) {
		if constexpr (emper::STATS) {
			auto flag = static_cast<unsigned int>(LawsStrategy::FiberSource::mainThread);
			fiber->setFlag(flag);
		}

		return fiber;
	}

	return nullptr;
}
