// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020 Florian Schmaus
#include <chrono>		 // for nanoseconds, duration_cast
#include <iostream>	 // for operator<<, basic_ostream, endl
#include <mutex>		 // for mutex
#include <thread>		 // for thread

#include "BinaryPrivateSemaphore.hpp"	 // for BPS
#include "Fiber.hpp"									 // for Fiber
#include "Runtime.hpp"								 // for Runtime
#include "emper-common.h"							 // for UNUSED_ARG

static auto threadTimeToSpawn() -> std::chrono::nanoseconds {
	std::chrono::time_point<std::chrono::high_resolution_clock> end;
	auto start = std::chrono::high_resolution_clock::now();
	std::thread t([&end] { end = std::chrono::high_resolution_clock::now(); });
	t.join();

	return std::chrono::duration_cast<std::chrono::nanoseconds>(end - start);
}

static auto fiberTimeToSpawn() -> std::chrono::nanoseconds {
	std::chrono::nanoseconds res;
	Runtime runtime;

	std::mutex testFinished;
	testFinished.lock();

	Fiber* fiber = Fiber::from([&runtime, &res, &testFinished] {
		BPS bps;
		std::chrono::time_point<std::chrono::high_resolution_clock> end;
		Fiber* fiber = Fiber::from([&end, &bps] {
			end = std::chrono::high_resolution_clock::now();
			bps.signal();
		});

		auto start = std::chrono::high_resolution_clock::now();
		runtime.schedule(*fiber);

		bps.wait();
		res = std::chrono::duration_cast<std::chrono::nanoseconds>(end - start);
		testFinished.unlock();
	});
	runtime.schedule(*fiber);

	testFinished.lock();

	// The destructor of Runtime causes all runtime threads to terminate.
	return res;
}

auto main(UNUSED_ARG int argc, UNUSED_ARG char* argv[]) -> int {
	{
		auto ttts = threadTimeToSpawn();
		std::cout << "Thread time to spawn: " << ttts.count() << " ns" << std::endl;
	}

	{
		auto ftts = fiberTimeToSpawn();
		std::cout << "Fiber time to spawn: " << ftts.count() << " ns" << std::endl;
	}
}
